def to_bytes(headers):
    """convert headers to canonical bytes array for signing"""
    values = values_sorted_by_key(headers)
    values_str = '\n'.join(values) + '\n'
    return values_str.encode()


# helpers helpers


def values_sorted_by_key(_dict):
    """returns list of values of _dict sorted by key"""
    return [_dict[key] for key in sorted(_dict.keys())]



from .cache_timer import CacheTimer
from .helpers     import to_bytes
from .sign        import sign


# constants
HEADER_VALID_FOR_S = 180
WIGGLE_ROOM_S      = 5
HEADER_LIFETIME_S  = HEADER_VALID_FOR_S - WIGGLE_ROOM_S


class SignedHeadersFactory:
    def __init__(self, consumer_id, private_key, key_version, passphrase=None):
        # client info
        self.consumer_id = consumer_id
        self.private_key = private_key
        self.key_version = key_version
        self.passphrase  = passphrase

        # lifecycle
        self.last_header = None
        self.cache_timer = CacheTimer(expires_in_s=HEADER_LIFETIME_S)

    def get_headers(self):
        if not self.cache_timer.is_valid:
            self.last_header = self.new_header()
        return self.last_header

    # helpers

    def new_header(self):
        # new header
        headers = {
            'WM_CONSUMER.ID':          self.consumer_id,
            'WM_SEC.KEY_VERSION':      self.key_version,
            'WM_CONSUMER.INTIMESTAMP': self.cache_timer.new_timestamp_str()
        }

        # sign
        data      = to_bytes(headers)
        signature = sign(self.private_key, data, self.passphrase)
        headers['WM_SEC.AUTH_SIGNATURE'] = signature
        return headers



from .helpers import new_keys
from pathlib  import Path


def generate_keys(secrets_path, passphrase=None):
    # is valid secrets path?
    secrets_path = Path(secrets_path)
    if not secrets_path.exists():
        raise ValueError(f"error: path '{secrets_path}' does not exist")
    if not secrets_path.is_dir():
        raise ValueError(f"error: path '{secrets_path}' is not a directory")

    # get key file paths
    public_key_file     = secrets_path / 'public_key.pem'
    public_key_b64_file = secrets_path / 'public_key.b64.txt'
    private_key_file    = secrets_path / 'private_key.pem'

    # generate new keys
    public_key, private_key = new_keys(passphrase)

    # save .pem files
    with public_key_file.open('wb') as f:
        f.write(public_key)
    with private_key_file.open('wb') as f:
        f.write(private_key)

    # save public_key.b64.txt
    public_key_b64 = public_key.lstrip(b'-----BEGIN PUBLIC KEY-----\n')
    public_key_b64 = public_key_b64.rstrip(b'-----END PUBLIC KEY-----\n')
    with public_key_b64_file.open('wb') as f:
        f.write(public_key_b64)



def use_mock_cache_timer(time_ms):
    # monkeypatch MockCacheTimer
    # into SignedHeadersFactory() class module
    import walmart_auth.signed_headers_factory.signed_headers_factory as shf_module
    MockCacheTimer        = get_mock_cache_timer_class(time_ms)
    shf_module.CacheTimer = MockCacheTimer


def get_mock_cache_timer_class(time_ms):
    class MockCacheTimer:
        def __init__(self, *args, **kwargs):
            # ignore all args
            self.is_valid = False

        def new_timestamp_str(self):
            # unix time since epoch
            # units: ms
            # e.g. '1234567890123'
            return str(time_ms)
    return MockCacheTimer



from argparse     import ArgumentParser
from walmart_auth import generate_keys


def main():
    # command line interface (CLI)
    parser = ArgumentParser(description='generate keys for use with requests-walmart-auth')
    parser.add_argument('secrets_path')
    parser.add_argument('passphrase', nargs='?', default=None)

    # parse CLI
    args         = parser.parse_args()
    secrets_path = args.secrets_path
    passphrase   = args.passphrase

    # generate keys
    generate_keys(secrets_path, passphrase)



from base64           import b64decode, b64encode
from Crypto.Hash      import SHA256
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5


def sign(private_key, data, passphrase=None):
    """return base64 hash (SHA256) and sign (PKCS1_v1_5) with private_key"""
    # hash, sign

    private_key = RSA.import_key(private_key, passphrase)
    signer    = PKCS1_v1_5.new(private_key)
    hasher    = SHA256.new(data)
    signature = signer.sign(hasher)

    # return as base64-encoded str
    return b64encode(signature).decode()



from .helpers     import interact_with_code
from argparse     import ArgumentParser
from walmart_auth import WalmartSession


def main():
    # command line interface (CLI)
    parser = ArgumentParser(description='Start Walmart requests session')
    parser.add_argument('--passphrase', '-p', nargs='?', default=None)
    parser.add_argument('consumer_id')
    parser.add_argument('private_key')
    parser.add_argument('key_version')


    # parse
    args         = parser.parse_args()
    consumer_id  = args.consumer_id
    private_key  = args.private_key
    passphrase   = args.passphrase
    key_version  = args.key_version

    if is_file(private_key):
        private_key = read_key_file(private_key)

    # open session
    session = WalmartSession(consumer_id, private_key, key_version, passphrase)

    # provide user REPL to
    # interact with session
    locals  = {
        'consumer_id': consumer_id,
        'private_key': private_key,
        'passphrase':  passphrase,
        'key_version': key_version,
        'session':     session,
    }
    interact_with_code(locals)



from time import time


class CacheTimer:
    @staticmethod
    def time_since_epoch_s():
        return time()

    def __init__(self, expires_in_s):
        self.expires_in_s = expires_in_s
        self.last_time_s  = None

    @property
    def is_valid(self):
        if self.last_time_s is None:
            return False
        # else
        return CacheTimer.time_since_epoch_s() <= self.expires_at_s

    @property
    def expires_at_s(self):
        return self.last_time_s + self.expires_in_s

    def new_timestamp_str(self):
        """returns str containing (rounded) time in milliseconds since epoch"""
        self.last_time_s = CacheTimer.time_since_epoch_s()
        return str(round(self.last_time_s * 1000))



from .signed_headers_factory import SignedHeadersFactory
from requests.auth           import AuthBase


class WalmartAuth(AuthBase):
    def __init__(self, consumer_id, private_key, key_version, passphrase=None):
        self.factory = SignedHeadersFactory(consumer_id, private_key, key_version, passphrase=None)

    def __call__(self, request):
        signed_headers = self.factory.get_headers()
        # import pdb; pdb.set_trace()
        request.headers.update(signed_headers)
        return request



from Crypto.PublicKey import RSA


# constants
BITS = 2048


def new_keys(passphrase=None):
    keys        = RSA.generate(BITS)
    public_key  = keys.publickey().export_key()
    private_key = keys.export_key(passphrase=passphrase, pkcs=8)
    return public_key, private_key



from setuptools import find_packages, setup


setup(
    name='requests-walmart-auth',
    version='1.0.0',
    description='python requests support for the Walmart Affiliate Marketplace (REST) API required Additional Headers',
    url='https://github.com/Terrabits/requests-walmart-auth',
    author='Nick Lalic',
    author_email='nick.lalic@gmail.com',
    license='MIT',
    packages=find_packages(exclude=['tests']),
    python_requires='>=3.7',
    install_requires=['pycryptodome~=3.14.1', 'requests~=2.27.1'],
    extras_require={
        'dev':  ['twine', 'wheel'],
    },
    entry_points={
        'console_scripts': [
            'generate-walmart-auth-keys=walmart_auth.bin.generate_keys:main',
            'walmart-session=walmart_auth.bin.session:main'
        ]
    }
)



import code
import os
from   pathlib import Path
import sys


def interact_with_code(local):
    # include current directory
    # in python import paths
    current_dir = os.getcwd()
    sys.path.insert(0, current_dir)

    # interact
    message = ''
    code.interact(message, local=local)


def is_file(file):
    file = Path(file)
    if not file.exists():
        return False
    return file.is_file()


def read_key_file(file):
    assert is_file(file)
    file = Path(file)
    with file.open() as f:
        return f.read()



from .auth    import WalmartAuth
from requests import Session


class WalmartSession(Session):
    def __init__(self, consumer_id, private_key, key_version, passphrase=None):
        Session.__init__(self)
        self.auth = WalmartAuth(consumer_id, private_key, key_version, passphrase)
